/* pfbtopfa.c
 *
 * A simple program to convert PostScript Type 1 fonts from the
 * binary format used on DOS and Windows (and perhaps other Intel
 * based) platforms to the ASCII PostScript format that can be
 * downloaded to any PostScript printer or interpreter regardless
 * of the underlying operating system.
 *
 * The PFB files contain a Type 1 PostScript font divided into
 * text and binary sections. Each section starts with a single byte
 * of 80 hexadecimal followed by another byte which is encoded
 * as follows:
 *
 *		1 - start of ASCII section
 *		2 - start of binary section
 *		3 - end of file
 *
 * The first two of these are followed by a 32-bit count of
 * the number of the data bytes in the section. This 32-bit
 * value is stored in the low-endian format.
 *
 * The data of the ASCII section can end its lines in one of
 * four possible methods, either by a CR or by a LF or CR/LF or
 * LF/CR. Most PFB fonts seem to end their lines with just a CR.
 * The binary data has no lines, so there is no concern with
 * line endings.
 *
 * The PFA format contains pure ASCII text. It still contains the
 * binary data but expressed as a series of hexadecimal digits,
 * two digits representing a single binary byte. It may be divided
 * into lines by inserting line endings, though it is not strictly
 * required. Since, however, some data communications channels might
 * be confused with a long line of ASCII data, this program adds
 * a line separator after each 32 bytes of data, i.e. after each 64
 * hexadecimal digits. This program also converts the line
 * endings of the originally ASCII section into a single new line
 * (which is a LF if compiled under Unix or CR/LF if compiled
 * under DOS/Windows).
 *
 * Written on 2003-02-06.
 *
 * Copyright 2003 G. Adam Stanislav.
 * All rights reserved.
 */

#include <stdio.h>

static void ordinary(int);
static void cr(int);
static void lf(int);

void (*outputchar)(int);

static FILE *input  = NULL;
static FILE *output = NULL;

static int usage(void) {
	fprintf(stderr,
		"pfbtopfa usage: pfbtopfa [[-i] pfbfile] [[-o] pfafile]\n"
	);
	return 1;
}

static int unexpected(void) {
	fprintf(stderr,
		"pfbtopfa: Unexpected end of input\n"
	);
	return 4;
}

static int invalid(void) {
	fprintf(stderr,
		"pfbopfa: Not a PFB input\n"
	);
	return 5;
}

static unsigned int getcount(void) {
	register unsigned int count;

	count  = (unsigned int)fgetc(input);
	count |= (unsigned int)fgetc(input) << 8;
	count |= (unsigned int)fgetc(input) << 16;
	count |= (unsigned int)fgetc(input) << 24;
	return count;
}

static void ordinary(int c) {
	switch(c) {
	case '\r':
		fputc('\n', output);
		outputchar = cr;
		break;
	case '\n':
		fputc('\n', output);
		outputchar = lf;
		break;
	default:
		fputc(c, output);
		break;
	}
}

static void cr(int c) {
	switch(c) {
	case '\r':
		fputc('\n', output);
		break;
	default:
		fputc(c, output);
	case '\n':
		outputchar = ordinary;
		break;
	}
}

static void lf(int c) {
	switch (c) {
	case '\n':
		fputc('\n', output);
		break;
	default:
		fputc(c, output);
	case '\r':
		outputchar = ordinary;
		break;
	}
}

static int ascii(void) {
	register unsigned int count;
	register int c;

	outputchar = ordinary;

	for (count = getcount(); count; count--) {
		if ((c = fgetc(input)) == EOF) return 1;
		outputchar(c);
	}
	return 0;
}

static int binary(void) {
	register unsigned int count;
	register int c, i;

	for (count = getcount(), i = 0; count; count--) {
		if ((c = fgetc(input)) == EOF) return 1;
		fprintf(output, "%.2x", c);
		if (++i == 32) {
			fputc('\n', output);
			i = 0;
		}
	}

	if (i) fputc('\n', output);

	return 0;
}

int main(int argc, char *argv[]) {
	register int i;
	register char *ptr;
	char *infile = NULL, *outfile = NULL;

	for (i = 1; i < argc; i++) {
		ptr = argv[i];

		if (ptr && *ptr) {
			if (*ptr == '-') {
				ptr++;
				switch (*ptr) {
				case 'i':
					ptr++;
					if (!*ptr) ptr = argv[++i];
					if (!ptr || !*ptr) return usage();
					infile = ptr;
					break;
				case 'o':
					ptr++;
					if (!*ptr) ptr = argv[++i];
					if (!ptr || !*ptr) return usage();
					outfile = ptr;
					break;
				default:
					return usage();
					break;
				}
			}
			else if (!infile)  infile  = ptr;
			else if (!outfile) outfile = ptr;
			else return usage();
		}
		else return usage();
	}

	if (infile) {
		if ((input = fopen(infile, "rb")) == NULL) {
			perror("pfbtopfa (input)");
			return 2;
		}
	}
	else input = stdin;

	if (outfile) {
		if ((output = fopen(outfile, "w")) == NULL) {
			perror("pfbtopfa (output)");
			if (input != stdin) fclose(input);
			return 3;
		}
	}
	else output = stdout;

	for (;;) {
		if ((i = fgetc(input)) == EOF)
			return unexpected();
		else if (i != 0x80)
			return invalid();

		switch (fgetc(input)) {
		case EOF:
			return unexpected();
			break;
		case 1:
			if (ascii()) return unexpected();
			break;
		case 2:
			if (binary()) return unexpected();
			break;
		case 3:
			if (input  != stdin)  fclose(input);
			if (output != stdout) fclose(output);
			return 0;
			break;
		default:
			return invalid();
		}
	}
}
